AUI_ANIMATION_VERTICAL = 101
AUI_ANIMATION_HORIZONTAL = 102
AUI_ANIMATION_ECLIPSE = 103

AUI_ANIMATION_MODE_FORWARD = 1000
AUI_ANIMATION_MODE_REVERSE_FORWARD = 1001
AUI_ANIMATION_MODE_BACKWARD = 1002
AUI_ANIMATION_MODE_REVERSE_BACKWARD = 1003

AUI.Animations = {}

local tasks = {}
local animationList = {}
local lastAnimations = {}
local animationCount = 0

local function GetAnimationCount(_animationList)
	local count = 0
	for id, _ in pairs(_animationList) do
		count = count + 1
	end

	return count
end

local function GetAnimationList(_parentName)	
	if not animationList[_parentName] then
		animationList[_parentName] = {}
	end
	
	return animationList[_parentName]
end

local function CreateNewAnimationControl(_virtualControl, _parentName, _parent)
	local controls = GetAnimationList(_parentName)	
	local controlId = GetAnimationCount(controls)
	local control = controls[controlId]

	if not control then
		control = CreateControlFromVirtual(_parentName .. "_Animation", _parent, _virtualControl, controlId)
		if not controls[controlId] then
			controls[controlId] = control
		end
	end	

	return control
end

local function GetAnimationControl(_virtualControl, _parent, _isVirtual)
	local parentName = _parent:GetName()
	
	if parentName then
		local controls = GetAnimationList(parentName)	
		for id, control in pairs(animationList[parentName]) do
			if not control.animation then
				control.animation = {}
			end
	
			if not control.animation.isPlaying then	
				return control
			end
		end
		
		local control = _virtualControl
		if _isVirtual then	
			control = CreateNewAnimationControl(_virtualControl, parentName, _parent, _isVirtual)
		end
		
		if not control.animation then
			control.animation = {}
		end
		
		return control
	end
	
	return nil
end

local function StopAnimation(_control)
	if _control and _control.animation and _control.animation.name then
		_control:SetAlpha(0)
		EVENT_MANAGER:UnregisterForUpdate("AUI_UPDATE_" .. _control.animation.name)
		_control.animation.isPlaying = false
	end
end

local function StartAnimation(_virtualControl, _parent, _duration, _animationType, _animationMode, _alignment, _startX, _startY, _endX, _endY, _isVirtual, _callback)
	local control = GetAnimationControl(_virtualControl, _parent, _isVirtual)		
	local animationName = control:GetName()

	control.animation.name = animationName
	control.animation.isPlaying = true
	control.animation.offsetX = 0
	control.animation.offsetY = 0	
	control.animation.ms = GetGameTimeMilliseconds()

	EVENT_MANAGER:RegisterForUpdate("AUI_UPDATE_" .. animationName, 0,
	function()
		if control.animation then
			local ms = GetGameTimeMilliseconds()
			local remaining = (ms - control.animation.ms) / _duration
		
			local startX = _startX or 0	          
			local startY = _startY or 0	
			local endX = _endX or 0	          
			local endY = _endY or 0			
			
			if _alignment == AUI_ALIGNMENT_CENTER then
				if _animationMode == AUI_ANIMATION_MODE_FORWARD or _animationMode == AUI_ANIMATION_MODE_BACKWARD then
					startX = _parent:GetWidth() / 2
					startY = _parent:GetHeight() / 2
				elseif _animationMode == AUI_ANIMATION_MODE_REVERSE_FORWARD or _animationMode == AUI_ANIMATION_MODE_REVERSE_BACKWARD then				
					endX = _parent:GetWidth() / 2 + endX
					endY = _parent:GetHeight() / 2 + endY		
				end				
			elseif _alignment == AUI_ALIGNMENT_LEFT then
				startX = 0	
			elseif _alignment == AUI_ALIGNMENT_RIGHT then
				startX = _parent:GetWidth()			
			elseif _alignment == AUI_ALIGNMENT_TOP then
				startY = 0	
			elseif _alignment == AUI_ALIGNMENT_BOTTOM then
				startY = _parent:GetHeight() / 2				
			elseif _alignment == AUI_ALIGNMENT_TOPLEFT then
				startX = 0	
				startY = 0
			elseif _alignment == AUI_ALIGNMENT_TOPRIGHT then
				startX = _parent:GetWidth()	
				startY = 0			
			elseif _alignment == AUI_ALIGNMENT_BOTTOMLEFT then
				startX = 0
				startY =  _parent:GetHeight()	
			elseif _alignment == AUI_ALIGNMENT_BOTTOMRIGHT then
				startX = _parent:GetWidth()
				startY =  _parent:GetHeight()
			end
					
			local parentWidth = _parent:GetWidth() - (startX + endX)
			local parentHeight = _parent:GetHeight() - (startY + endY)
			local offsetX = 0	          
			local offsetY = 0			
			local alpha = _startAlpha or 0
			local point = TOP
			local rPoint = TOP	
	
			if _animationMode == AUI_ANIMATION_MODE_FORWARD then
				point = TOPLEFT
				rPoint = TOPLEFT		
		
				if _animationType == AUI_ANIMATION_ECLIPSE then
					offsetX = (-parentWidth * ((2 * (remaining * remaining)) - (2 * remaining))) + startX
					offsetY = ((parentHeight * remaining) + startY)
				
					point = TOPLEFT
					rPoint = TOPLEFT					
				elseif _animationType == AUI_ANIMATION_VERTICAL then
					offsetY = (parentHeight * remaining + startY)
				
					point = TOP
					rPoint = TOP					
				elseif _animationType == AUI_ANIMATION_HORIZONTAL then
					offsetX = (parentWidth * remaining) + startX
				
					point = LEFT
					rPoint = LEFT						
				end					
			elseif _animationMode == AUI_ANIMATION_MODE_REVERSE_FORWARD then
				if _animationType == AUI_ANIMATION_ECLIPSE then
					offsetX = (parentWidth * ((2 * (remaining * remaining)) - (2 * remaining))) - startX
					offsetY = (parentHeight * remaining - startY)
				
					point = TOP
					rPoint = TOP					
				elseif _animationType == AUI_ANIMATION_VERTICAL then
					offsetY = (parentHeight * remaining) + startY
				
					point = TOP
					rPoint = TOP					
				elseif _animationType == AUI_ANIMATION_HORIZONTAL then
					offsetX = (parentWidth * remaining) + startX	
				
					point = LEFT
					rPoint = LEFT					
				end				
			elseif _animationMode == AUI_ANIMATION_MODE_BACKWARD then
				if _animationType == AUI_ANIMATION_ECLIPSE then
					offsetX = (-parentWidth * ((2 * (remaining * remaining)) - (2 * remaining))) - startX
					offsetY = (-parentHeight * remaining) - startY

					point = BOTTOMLEFT
					rPoint = BOTTOMLEFT
				elseif _animationType == AUI_ANIMATION_VERTICAL then
					offsetY = (-parentHeight * remaining) - startY
			
					point = BOTTOM
					rPoint = BOTTOM	
				elseif _animationType == AUI_ANIMATION_HORIZONTAL then				
					offsetX = -(parentWidth * remaining) - startX		
				
					point = RIGHT
					rPoint = RIGHT					
				end					
			elseif _animationMode == AUI_ANIMATION_MODE_REVERSE_BACKWARD then
				if _animationType == AUI_ANIMATION_ECLIPSE then
					offsetX = (parentWidth * ((2 * (remaining * remaining)) - (2 * remaining))) - startX
					offsetY = (-parentHeight * remaining) - startY
			
					point = BOTTOM
					rPoint = BOTTOM	
				elseif _animationType == AUI_ANIMATION_VERTICAL then 
					offsetY = (-parentHeight * remaining) - startY
			
					point = BOTTOM
					rPoint = BOTTOM	
				elseif _animationType == AUI_ANIMATION_HORIZONTAL then
					offsetX = (-parentWidth * remaining) - startX
				
					point = RIGHT
					rPoint = RIGHT					
				end			
			end
			
			control.animation.offsetX = offsetX
			control.animation.offsetY = offsetY
			control.animation.width = control:GetWidth()
			control.animation.height = control:GetHeight()
			control.animation.animationType = _animationType
			control.animation.animationMode = _animationMode
			
			if _callback then
				_callback(control)
			end	
			
			if remaining > 0 then
				alpha = -((remaining * remaining) - remaining) * 4
			end			
			
			if alpha <= 0 then
				StopAnimation(control)	
			end
		
			control:ClearAnchors()
			control:SetAnchor(point, _parent, rPoint, offsetX, offsetY)
			control:SetAlpha(alpha)	
		else
			EVENT_MANAGER:UnregisterForUpdate("AUI_UPDATE_" .. animationName)
		end
	end)
	
	return control.animation
end

local function IsAnimationApproved(_taskData)
	local panelName = _taskData.parent:GetName()
	if panelName then
		local lastAnimation = lastAnimations[panelName]	
		if lastAnimation then
			local currentMS = GetGameTimeMilliseconds() - lastAnimation.ms	
			local interval = (_taskData.duration / 1000) * 90
			if currentMS >= interval then
				return true	
			end
		else
			return true
		end
	end
	
	return false
end

local function ExecuteTasks()
	if animationCount > 0 then
		EVENT_MANAGER:UnregisterForUpdate("AUI_UPDATE_SCROLLING_TEXT_TASKS")	

		for taskId, taskData in pairs(tasks) do		
			local panelName = taskData.parent:GetName()	
			if panelName then	
				if IsAnimationApproved(taskData) then
					lastAnimations[panelName] = StartAnimation(taskData.virtualControl, taskData.parent, taskData.duration, taskData.animationType, taskData.animationMode, taskData.alignment, taskData.startX, taskData.startY, taskData.endX, taskData.endY, taskData.isVirtual, taskData.callback)
					
					tasks[taskId] = nil	
					animationCount = animationCount - 1				
				end
			end
		end
		
		EVENT_MANAGER:RegisterForUpdate("AUI_UPDATE_SCROLLING_TEXT_TASKS", 100, ExecuteTasks)
	end
end

function AUI.Animations.RemoveFrom(_panel)
	if not _panel then
		return
	end

	local panelName = _panel:GetName()
	
	for taskId, taskData in pairs(tasks) do	
		local taskDataParentName = taskData.parent:GetName()
		if taskDataParentName == panelName then
			tasks[taskId] = nil
			animationCount = animationCount - 1
		end
	end
	
	lastAnimations[panelName] = nil
	
	if animationList[panelName] then
		for id, control in pairs(animationList[panelName]) do	
			control.animation = nil
			control:SetAlpha(0)
		end
	end
end

function AUI.Animations.IsPlaying(_control)
	if _control and _control.animation then
		return _control.animation.isPlaying
	end
end

function AUI.Animations.StopAnimation(_control)
	if _control and _control.animation then
		StopAnimation(_control)	
	end
end

function AUI.Animations.StartAnimation(_virtualControl, _parent, _duration, _animationType, _animationMode, _alignment, _startX, _startY, _endX, _endY, _isVirtual, _callback)
	if not _virtualControl or not _parent or not _duration or not _animationType or not _animationMode then
		return
	end
	
	if _isVirtual then
		table.insert(tasks, {virtualControl = _virtualControl, parent = _parent, duration = _duration or 1, animationType = _animationType, animationMode = _animationMode, alignment = _alignment, isVirtual = _isVirtual, startX = _startX, startY = _startY, endX = _endX, _endY = _endY, callback = _callback})
		animationCount = animationCount + 1	
	else
		StartAnimation(_virtualControl, _parent, _duration, _animationType, _animationMode, _alignment, _startX, _startY, _endX, _endY, _isVirtual, _callback)
	end
end

EVENT_MANAGER:RegisterForUpdate("AUI_UPDATE_SCROLLING_TEXT_TASKS", 100, ExecuteTasks)